<?php

namespace Drupal\varnish_purge_tags_override\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\ConfigFormBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\varnish_purge_tags_override\VarnishPurgeTagsOverrideHelperInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Configure varnish_purge_tags_override settings for this site.
 *
 * @internal
 */
class SettingsForm extends ConfigFormBase
{

    /**
     * @var \Drupal\varnish_purge_tags_override\VarnishPurgeTagsOverrideHelperInterface
     */
    protected $helper;

    /**
     * Constructs a new SettingsForm.
     *
     * @param \Drupal\varnish_purge_tags_override\VarnishPurgeTagsOverrideHelperInterface $helper
     *
     * @param \Drupal\varnish_purge_tags_override\VarnishPurgeTagsOverrideHelperInterface $helper
     */
    public function __construct(ConfigFactoryInterface $configFactory, VarnishPurgeTagsOverrideHelperInterface $helper)
    {
        parent::__construct($configFactory);
        $this->helper = $helper;
    }

    /**
     * {@inheritdoc}
     */
    public static function create(ContainerInterface $container)
    {
        return new static(
            $container->get('config.factory'),
            $container->get('varnish_purge_tags_override.helper')
        );
    }

    /**
     * {@inheritdoc}
     */
    public function getFormId()
    {
        return 'varnish_purge_tags_override_settings_form';
    }

    /**
     * {@inheritdoc}
     */
    protected function getEditableConfigNames()
    {
        return ['varnish_purge_tags_override.settings'];
    }

    /**
     * {@inheritdoc}
     */
    public function buildForm(array $form, FormStateInterface $form_state)
    {
        $config = $this->config('varnish_purge_tags_override.settings');
        $maximum_header_value_size = $config->get('maximum_header_value_size');
        if (null === $maximum_header_value_size) {
            $maximum_header_value_size = VarnishPurgeTagsOverrideHelperInterface::DEFAULT_MAXIMUM_HEADER_VALUE_SIZE;
        }
        $description = new TranslatableMarkup('Maximum size in bytes of the Cache-Tags header value<br/ >Default size is : @default_size', ['@default_size' => VarnishPurgeTagsOverrideHelperInterface::DEFAULT_MAXIMUM_HEADER_VALUE_SIZE]);
        $form['maximum_header_value_size'] = [
            '#type' => 'number',
            '#title' => $this->t('Maximum size in bytes of the Cache-Tags header value'),
            '#default_value' => $maximum_header_value_size,
            '#description' => $description,
            "#min" => 1,
        ];

        $pages =  $config->get('pages');
        if (null === $pages) {
            $pages = "";
        }
        $form['pages'] = [
            '#type' => 'textarea',
            '#title' => $this->t('Pages'),
            '#default_value' => $pages,
            '#description' => $this->t(
                "Specify pages by using their paths. Enter one path per line. The '*' character is a wildcard. An example path is %user-wildcard for every user page. %front is the front page.", [
                '%user-wildcard' => '/user/*',
                '%front' => '<front>',
                ]
            ),
        ];

        $types =  $config->get('types');
        if (null === $types) {
            $types = ["node" => "node"];
        }

        $contentEntityTypes = $this->helper->listContentEntityTypes();
        $form['content_entity_types'] = [
            '#type' => 'checkboxes',
            '#title' => $this->t('Content entity types'),
            '#options' => $contentEntityTypes,
            '#default_value' => $types,
            '#description' => $this->t("Specify which content entity types are treated by this module when altering the cache tags."),
        ];

        $per_bundle =  $config->get('per_bundle');
        if (null === $per_bundle) {
            $per_bundle = false;
        }
        $form["options"] = [
            '#type' => 'fieldset',
            '#title' => $this->t('Advanced options'),
            '#collapsible' => true,
            '#collapsed' => false,
        ];
        $form["options"]["per_bundle"] = [
            '#type' => 'checkbox',
            '#title' => $this->t('Per bundle'),
            '#default_value' => $per_bundle,
            '#description' => $this->t("If checked, the module will alter the cache tags for each bundle of each content entity type and add the list tag for each bundle type instead of the list tag for entire entity type."),
        ];

        return parent::buildForm($form, $form_state);
    }

    public function validateForm(array &$form, FormStateInterface $form_state)
    {
        parent::validateForm($form, $form_state); // TODO: Change the autogenerated stub
        $paths = array_map('trim', explode("\n", $form_state->getValue('pages')));
        foreach ($paths as $path) {
            if (empty($path) || $path === '<front>' || str_starts_with($path, '/')) {
                continue;
            }
            $form_state->setErrorByName('pages', $this->t("The path %path requires a leading forward slash when used with the Pages setting.", ['%path' => $path]));
        }
    }

    /**
     * {@inheritdoc}
     */
    public function submitForm(array &$form, FormStateInterface $form_state)
    {

        $types = $form_state->getValue('content_entity_types');
        $types = array_filter(
            $types, function ($value, $key) {
                return ($value !== null) && ($value !== false) && ($value !== '') && ($value !== 0) && ($value !== '0');
            }, ARRAY_FILTER_USE_BOTH
        );
        $types = array_keys($types);

        $this->config('varnish_purge_tags_override.settings')
            ->set('maximum_header_value_size', $form_state->getValue('maximum_header_value_size'))
            ->set('pages', $form_state->getValue('pages'))
            ->set('per_bundle', $form_state->getValue('per_bundle'))
            ->set('types', $types)
            ->save();

        parent::submitForm($form, $form_state);
    }


}
